from .type import DefaultEntryTypes, ComplexData, ComplexEntryType

import logging
logger = logging.getLogger('nt')

__all__ = ["NetworkTableNode"]

class NetworkTableNode:
    """represents a node (either a client or a server) in a network tables 2.0
    """

    def __init__(self, entryStore):
        self.entryStore = entryStore
        self.remoteListeners = []
        self.tableListeners = []

    def getEntryStore(self):
        """:returns: the entry store used by this node
        """
        return self.entryStore

    def putBoolean(self, name, value):
        self.putValue(name, value, type=DefaultEntryTypes.BOOLEAN)

    def getBoolean(self, name):
        entry = self.entryStore.getEntry(name)
        if entry is None:
            raise KeyError(name)
        if entry.getType() != DefaultEntryTypes.BOOLEAN:
            raise TypeError("Cannot get boolean for '%s', is a %s" % (name, entry.getType().name))
        return entry.getValue()

    def putNumber(self, name, value):
        self.putValue(name, value, type=DefaultEntryTypes.DOUBLE)

    def getNumber(self, name):
        entry = self.entryStore.getEntry(name)
        if entry is None:
            raise KeyError(name)
        if entry.getType() != DefaultEntryTypes.DOUBLE:
            raise TypeError("Cannot get number for '%s', is a %s" % (name, entry.getType().name))
        return entry.getValue()

    putDouble = putNumber
    getDouble = getNumber


    def putString(self, name, value):
        self.putValue(name, value, type=DefaultEntryTypes.STRING)

    def getString(self, name):
        entry = self.entryStore.getEntry(name)
        if entry is None:
            raise KeyError(name)
        if entry.getType() != DefaultEntryTypes.STRING:
            raise TypeError("Cannot get string for '%s', is a %s" % (name, entry.getType().name))
        return entry.getValue()

    def putComplex(self, name, value):
        self.putValue(name, value, type=value.getType())

    def retrieveValue(self, name, externalData):
        with self.entryStore.entry_lock:
            entry = self.entryStore.getEntry(name)
            if entry is None:
                raise KeyError(name)
            entryType = entry.getType()
            if not isinstance(entryType, ComplexEntryType):
                raise TypeError("Cannot get complex data for '%s', is a %s" % (name, entryType.name))
            entryType.exportValue(name, entry.getValue(), externalData)

    def putValue(self, name, value, type=None):
        """Put a value with a specific network table type
        :param name: the name of the entry to associate with the given value
        :param value: the actual value of the entry
        :param type: the type of the entry (if not provided, will be guessed)
        """
        if type is None:
            # Guess type based on value
            if isinstance(value, bool):
                type = DefaultEntryTypes.BOOLEAN
            elif isinstance(value, (float, int)):
                type = DefaultEntryTypes.DOUBLE
            elif isinstance(value, str):
                type = DefaultEntryTypes.STRING
            elif isinstance(value, ComplexData):
                type = value.getType()
            elif value is None:
                raise ValueError("Cannot put a null value into networktables")
            else:
                raise ValueError("Invalid Type")
        
        self.entryStore.putOutgoing(name, type, value)

    def getValue(self, name):
        #TODO don't allow get of complex types
        with self.entryStore.entry_lock:
            entry = self.entryStore.getEntry(name)
            if entry is None:
                raise KeyError(name)
            return entry.getValue()

    def containsKey(self, key):
        """:param key: the key to check for existence
        :returns: True if the table has the given key
        """
        return self.entryStore.getEntry(key) is not None

    def __contains__(self, key):
        return self.entryStore.getEntry(key) is not None

    def close(self):
        """close all networking activity related to this node
        """
        pass
    
    def getRemoteAddress(self):
        return None
    
    def isConnected(self):
        raise NotImplementedError()

    def addConnectionListener(self, listener, immediateNotify):
        self.remoteListeners.append(listener)
        if self.isConnected():
            listener.connected(self)
        else:
            listener.disconnected(self)

    def removeConnectionListener(self, listener):
        self.remoteListeners.remove(listener)

    def fireConnectedEvent(self):
        for listener in self.remoteListeners:
            listener.connected(self)

    def fireDisconnectedEvent(self):
        for listener in self.remoteListeners:
            listener.disconnected(self)

    def addTableListener(self, listener, immediateNotify):
        self.tableListeners.append(listener)
        if immediateNotify:
            self.entryStore.notifyEntries(None, listener)

    def removeTableListener(self, listener):
        self.tableListeners.remove(listener)

    def fireTableListeners(self, key, value, isNew):
        for listener in self.tableListeners:
            try:
                listener.valueChanged(None, key, value, isNew)
            except Exception:
                logger.exception('Exception in valueChanged callback!')




import socket
import sys

from . import _impl

__all__ = ["SocketStreamFactory", "SocketServerStreamProvider"]


class SocketStream:
    def __init__(self, conn):
        self.conn = conn

    def getInputStream(self):
        return _impl.sock_makefile(self.conn, 'rb')

    def getOutputStream(self):
        return _impl.sock_makefile(self.conn, 'wb')
    
    def getRemoteAddress(self):
        return self.conn.getpeername()[0]

    def close(self):
        self.conn.close()

class SocketStreamFactory:
    def __init__(self, host, port):
        self.host = host
        self.port = port

    def createStream(self):
        return SocketStream(_impl.sock_create_connection((self.host, self.port)))

class SocketServerStreamProvider:
    def __init__(self, port):
        self.server = socket.socket()
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind(('', port))
        self.server.listen(50)
        
        # this fails in 2.7... need a better way of dealing with this
        # -> However, robot servers are only running python 3, so it doesn't
        #    really matter at the moment..
        if sys.version_info[0] > 2:
            self.server.settimeout(1.0)

    def accept(self):
        return SocketStream(self.server.accept()[0])




from networktables2 import StringArray
from .networktable import NetworkTable


def ntproperty(key, defaultValue, writeDefault=True):
    '''
        A property that you can add to your classes to access NetworkTables
        variables like a normal variable.
        
        :param key: A full NetworkTables key (eg ``/SmartDashboard/foo``)
        :type  key: str
        :param defaultValue: Default value to use if not in the table
        :type  defaultValue: any
        :param writeDefault: If True, put the default value to the table,
                             overwriting existing values
        :type  writeDefault: bool
        
        Example usage::
        
            class Foo(object):
            
                something = ntproperty('/SmartDashboard/something', True)
                
                ...
                
                def do_thing(self):
                    if self.something:    # reads from value
                        ...
                        
                        self.something = False # writes value
        
        .. note:: Does not work with complex types.
        
                  Getting the value of this property should be reasonably
                  fast, but setting the value will have just as much overhead
                  as :meth:`.NetworkTable.putValue`
                  
        .. versionadded:: 2015.3.0
    '''
    ntvalue = NetworkTable.getGlobalAutoUpdateValue(key, defaultValue, writeDefault)
    nttable = []
    
    def _get(_):
        return ntvalue.value
    
    def _set(_, value):
        if not nttable:
            nttable.append(NetworkTable.getGlobalTable())
        nttable[0].putValue(key, value)
        ntvalue._AutoUpdateValue__value = value
    
    return property(fget=_get, fset=_set)



class ChooserControl(object):
    '''
        Interacts with a :class:`wpilib.sendablechooser.SendableChooser`
        object over NetworkTables.
    '''
    
    def __init__(self, key, on_choices=None, on_selected=None):
        '''
            :param key: NetworkTables key
            :type  key: str
            :param on_choices: A function that will be called when the
                               choices change. Signature: fn(value)
            :param on_selection: A function that will be called when the
                                 selection changes. Signature: fn(value)
        '''
        
        self.subtable = NetworkTable.getTable('SmartDashboard').getSubTable(key)

        self.on_choices = on_choices
        self.on_selected = on_selected
        
        if on_choices or on_selected:
            self.subtable.addTableListener(self._on_change, True)

    def close(self):
        '''Stops listening for changes to the ``SendableChooser``'''
        if self.on_choices or self.on_selected:
            self.subtable.removeTableListener(self._on_change)
    
    def getChoices(self):
        '''
            Returns the current choices. If the chooser doesn't exist, this
            will return an empty array.
        
            :rtype: :class:`.StringArray`
        '''
        choices = StringArray()
        try:
            self.subtable.retrieveValue('options', choices)
        except KeyError:
            pass
        return choices
    
    def getSelected(self):
        '''
            Returns the current selection or None
        
            :rtype: str
        '''
        selected = self.subtable.getString('selected', None)
        if selected is None:
            selected = self.subtable.getString('default', None)
        return selected
    
    def setSelected(self, selection):
        '''
            Sets the active selection on the chooser
            
            :param selection: Active selection name
        '''
        self.subtable.putString('selected', selection)
    
    def _on_change(self, table, key, value, isNew):
        if key == 'options':
            if self.on_choices is not None:
                self.on_choices(value)
        elif key == 'selected':
            if self.on_selected is not None:
                self.on_selected(value)
        elif key == 'default':
            if self.on_selected is not None and self.subtable.getString('selected', None) is None:
                self.on_selected(value)



#!/usr/bin/env python3
#
# This is a NetworkTables client (eg, the DriverStation/coprocessor side).
# You need to tell it the IP address of the NetworkTables server (the
# robot or simulator).
#
# This shows how to use a listener to listen for changes in NetworkTables
# values. This will print out any changes detected on the SmartDashboard
# table.
#

import sys
import time
from networktables import NetworkTable

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

if len(sys.argv) != 2:
    print("Error: specify an IP to connect to!")
    exit(0)

ip = sys.argv[1]

NetworkTable.setIPAddress(ip)
NetworkTable.setClientMode()
NetworkTable.initialize()

def valueChanged(table, key, value, isNew):
    print("valueChanged: key: '%s'; value: %s; isNew: %s" % (key, value, isNew))

class ConnectionListener:

    def connected(self, table):
        print("Connected to", table.getRemoteAddress(), table)

    def disconnected(self, table):
        print("Disconnected", table)

c_listener = ConnectionListener()

sd = NetworkTable.getTable("SmartDashboard")
sd.addTableListener(valueChanged)
sd.addConnectionListener(c_listener)

while True:
    time.sleep(1)

 
